puts(1, "See ternary3.let for info about WORDS.TXT -- Rob Craig\n\n")

--To run, type ex captree.ex

--Implementation of ternary tree per Dr Dobbs Journal, April, 1998, page 20
--This prog by Art Adamson, euclid@isoc.net...May 25, 1998
--Reference, http://www.cs.princeton.edu/~rs/strings/
--This version contains only the search and insert (tree build) functions.
--The input file format is words with a terminator, one word per line.

--This version reverse sorts the input words causing a degenerate tree..ie..
--all the node pointers are to left or center, none to right. Therefore there
--is no need for right storage...thought limited right storage capibiliy is 
--retained for future developments. The center pointer storage need is also
--reduced to a one-bit flag per node. Hence, within the 32 bits of an integer,
--there is ample room for a large tree. The acceptable characters are 
--currently 1..255 though I have checked only slightly outside of printable
--ascii (32..126) chars.

--This version, June 1, 1998, is probably slightly slower than a balanced tree
--but it is still fast, compact, and has a large capacity.

--This prog sorts the input words before storing them. The sort time is not 
--included in the announced store time. For a real dictionary, the input
--should be sorted, z first. Then the sort calls below could be commented out.
--The dictionary would be formed via the insert technique here-in and saved.
--It would then be loaded when desired for use, probably via Ralf's EDOM,
--especially if he can speed up the save portion.

--The first lines in capnodes.e define bits. They determine what is stored
--in each bit. As now set, 
--charBits   1..8 store the input characters, 0..255
--End of Word flag in bit 9
--left pointers in 10..29
--center pointer in 30
--right pointers in 31..32

--The left storage capacity is large (study an output print-out to understand).
--The right pointer storage capacity is only 3. In normal usage, 0 is 
--sufficient. The right storage is used only to signal trouble via an
--unexpected call for a right storage.

--This prog could serve as a basis for spell checking, dictionaries, etc.

--I haven't yet packaged it for easy use.
---------------------------------------------------------------------------
--A typical screen capture follows for this developmental program using 
--input from words.txt from the RDS spell Checker.

--Input the number of words...I will read them from words1.txt
--I will show more detail if you request fewer word/s than: 300
--Num of words is : 60000
--Input the maxWordLength: maxWordLength is: 25
--Words loaded! I selected from those eligible:  51802

--I will build the tree.  Tree built!!  Search-time check done!!  
--Search-completness check done!!       I found all the words!!!


--Store and search data follows :

--Store and search times                            :  36.4464  13.8539
--Average store and search times per word           :  0.000704 0.000267
--
--numWords & maxWordLength are                      :  51802   22
--Num of input-chars stored                         :  404584
--Average word length                               :  7.81
--
--Nodes used for storing the tree, limit : used     :  1048575   :  129881
--Nodes used per input char                         :  0.321022
--Total bytes for tree storage including pointers is:  519572
--Num of storage bytes per input char is            :  1.28421
--

--RNodes total limit : used                         :  3 : 0
--Bytes to store RPointers)                         :  24
--
--Bytes for inputWords Sequence, just for comparison:  2861616
--Compression ratio bytes(sequence) / bytes(nodes)  :  5.5079
--------------------------------------------------------------------------

include capinsrt.e
include capnodes.e
include capserch.e
include cprint.e
include ranWords.e  
include bytes.e
include get.e
include wildcards.e
include sort.e
--include edom.e
without type_check
without warning
--with trace
object line, wordsIn, word
object o1 , o2
atom t1, t2, t3, t4
sequence  words, inFileName, outFileName
integer numWords, maxWordLength, printLimit, charsIn, wordsInBytes, dupCount
integer findsCount, charCount, fIn, fOut, count
sequence outFileName1, outFileName2, outFileName3
--inFileName = "words.prp"
--inFileName = "10000.out"
--inFileName = "words.ori"
--inFileName = "wrds100k.in"
--inFileName = "1505666n.dup"
--inFileName = "29118nd.obj"
--inFileName = "mixedwrd.in"
--inFileName = "halfcaps.in"
--inFileName = "words.ori"
inFileName = "words.txt"
--inFileName = "words1.txt"
--inFileName = "12.out"
--inFileName = "wa.a"
--inFileName = "123801.asc"
--
--inFileName = "6500078n.obj"
--outFileName1 = "printree.out"
--outFileName2 = "edomtree.out"
outFileName3 = "usedwrds.out"

findsCount = 0
integer ptr
tick_rate(500)
printLimit = 300    --Sets limit for extended info output. Edit this if needed

--****************get user inputs
puts(1,
"Input the number of words...I will read them from " & inFileName)
puts(1,"\nI will show more detail if you request fewer word/s than: ") 
? printLimit
puts(1,"Num of words is :  ")
o1 = get(0)      numWords = o1[2]
puts(1,"\n")

puts(1,
"Input the maxWordLength: ")
o1 = get(0)      maxWordLength = o1[2]
puts(1,"maxWordLength is: ")   ? maxWordLength


findsCount = 0
wordsIn = {}
fIn = open(inFileName, "r")
fOut = open(outFileName3, "w")

--wordsIn = get(fIn)
--? length(wordsIn)
--wordsIn = wordsIn[2]

--for i = 1 length(wordsIn)  do
--for i = 43430 to 43450   do
--puts(fOut,wordsIn[i] & "\n")
--end for

--******************Load the words 
count = 0       charCount = 0
while count < numWords do
  word = gets(fIn)
  if atom(word) then  exit        
  end if
  if length(word) <= maxWordLength + 1 then
    count = count + 1
    wordsIn = append (wordsIn,word[1..length(word) - 1])
    charCount = charCount + length(word) - 1
  end if
  if count = numWords  then    exit
  end if
end while   
puts(1,"Words loaded! I selected from those eligible:  ") 
print(1,length(wordsIn))
numWords = length(wordsIn)
close(fIn)
--wordsIn[2] = {67,76,65,132,82,69}   --add a test word beyond ascii
--wordsIn = append(wordsIn,{67,76,65,132,82,69}) --add a test word beyond ascii

--************************find maxWordLength
maxWordLength = 0
for i = 1 to length(wordsIn) by 1 do
  if length(wordsIn[i]) > maxWordLength then
    maxWordLength = length(wordsIn[i])
  end if
end for

--**********************Count bytes for comparison reasons.
wordsInBytes = bytes(wordsIn)

--********************Show input words if not too many
if length(wordsIn) < printLimit then 
  puts(1,"\nThe input words are:\n")  cprint(wordsIn)
  puts(1,"\n")
end if

--****************Sort wordsIn, so we get an unbalanced tree
wordsIn = sort(wordsIn)

--*******************Reverse the order, so only left pointers are used
for i = 1 to floor(length(wordsIn)/2)  do
  word = wordsIn[i]
  wordsIn[i] = wordsIn[numWords -i+1]
  wordsIn[numWords -i+1] = word
end for

--output the words for a format check if needed
for i = 1 to length(wordsIn)  do
  puts(fOut, wordsIn[i] & "\n") 
end for

--Build the tree in a global sequence "nodes" in unodes.e. Check the time
puts(1,"\nI will build the tree.  ")
ptr = 0

--for i = 43430 to 43450   do
--puts(fOut,wordsIn[i] & "\n")
--end for
t1 = time()
for i = 1 to length(wordsIn)   do 
  ptr =  insert(ptr,(wordsIn[i] ),0)
end for
t2 = time()                         --t1-t2 = build time for numWords
puts(1,"Tree built!!  ")

--Check search routine, see if we can find all the words, check time only
t3 = time()
for i = 1 to length(wordsIn)  do   
  o1 =  search(wordsIn[i]) 
end for
t4 = time()
puts(1,"Search-time check done!!  ")

--Check ability to find every word
findsCount = 0
o1 = {}
--trace(1)
for i = 1 to  length(wordsIn)   do   
  if  search(wordsIn[i]) = 1 then --o1 = append(o1,wordsIn[i]) 
    findsCount = findsCount + 1  
--    else cprint(wordsIn[i]) ? i 
    --cprint(wordsIn[i-1])
    --cprint( wordsIn[i])
    --cprint( wordsIn[i+1])
  end if
end for
--trace(0)
puts(1,"\nSearch-completness check done!!  ")

if findsCount = numWords then puts(1, "     I found all the words!!!")
else puts(1,"I found only  ") print(1,findsCount)  puts(1," !!words???")
end if

--Repeat the find but output the resulting 0's or 1's. 1 means found
if length(wordsIn) < printLimit then    --enhanced output
  puts(1,"\nSearching for each original input word, I get:\n")
--  if length(wordsIn) < printLimit then 
    for i = 1 to length(wordsIn)  do
      print(1,search(wordsIn[i]))
    end for
  --end if  
  
  
  --puts(1,
  --"\nSearching, in reverse order, the input words; search still works: \n")
  --for i = length(wordsIn) to 1 by -1  do
  --print(1,search(wordsIn[i]))
  --end for
  
  --puts(1,"\nUsing only alternate input words, search still works: \n")
  --for i = 1 to length(wordsIn) by 2 do
  --print(1,search(wordsIn[i]))
  --end for
  
  --Check for rejection of different random-words
  puts(1,
  "\n\nSearch results for similar but different random words follow:\n")
  puts(1,
  "This should be a good way to compare two lists for common members.\n")
  puts(1,
  "The  1's in result represent commons, if any, in the 2 test word lists\n")
  
  --Release the memory, and generate a list of random-char words
  wordsIn = {}   
  randWordsIn = {}
  randWords(maxWordLength,numWords, 21000, "zymase")  --21000 random init
randWordsIn[2] = {67,76,65,132,82,69}
  puts(1,"\n")
  findsCount = 0
  o1 = {}      o2 = 0
  
  --Use the random word list to test ability of search to reject words not
  --stored in the tree.
  for i = 1 to length(randWordsIn)  do
    o2 =    search(randWordsIn[i]) 
    print(1,o2)
    if o2 = 1 then  findsCount = findsCount + 1
      o1 = append(o1, randWordsIn[i]) 
    end if
  end for
  --Some of the random words will be found to agree with the original words
  --as in monkeys typing
  puts(1,"\nThe common words are:\n")
  cprint(o1)
  puts(1,"\nI found, ")   
  print(1, findsCount) puts(1,"  words common to the 2 lists.\n")
  randWordsIn = {}      o1 = {}
  --end of under-print-limit actions.
  
  --Start of over-print-limit actions
else      --not enhanced output
  o1 = {}
  wordsIn = {}   
  randWordsIn = {}
  randWords(maxWordLength,numWords, 111, "zymase")
randWordsIn[2] = {67,76,65,132,82,69}
  --zymase is a word in the original word list for 13000 words, 6 max length
  puts(1,"\n")
  findsCount = 0
  o1 = {}      o2 = 0
  for i = 1 to length(randWordsIn)  do
  o2 =    search(randWordsIn[i]) 
  if o2 = 1 then  findsCount = findsCount + 1
  o1 = append(o1, randWordsIn[i]) 
  end if
  end for
  randWordsIn = {}
  
  puts(1,"\nI found, ")   
  print(1, findsCount) puts(1," common words in " & inFileName & 
  " and a random words file.\n")
  --puts(1," With these, I re-searched the tree to check.  ")  
  --puts(1,"\nI found these commons: \n ")   
  -----------------cprint(o1)  puts(1,"  OR  ")    ? o1
  --findsCount = 0
  --for i = 1 to length(o1)  do
  --if  search(o1[i]) 
  --then    findsCount = findsCount + 1
  --end if
  --end for
  --print(1, findsCount)
  --puts(1," words!!!\n") -- They really were common to the 2 lists.\n")    
end if


puts(1,"\n\nStore and search data follows :") 
puts(1,"\n\nStore and search times                            :  ") 
print(1,(t2-t1))  puts(1,"  ") print(1,(t4-t3))

puts(1,"\nAverage store and search times per word           :  ") 
printf(1,"%0.6f\t",(t2-t1)/numWords)
printf(1,"%0.6f\n",(t4-t3)/numWords)

puts(1,"\nnumWords & maxWordLength are                      :  ") 
print(1,numWords) 
puts(1,"   ")       ? maxWordLength

puts(1,"Num of input-chars stored                         :  ")  
? charCount

puts(1,"Average word length                               :  ")  
? charCount/numWords

puts(1,"\n")
puts(1,"Nodes used for storing the tree, limit : used     :  ")
print(1, power(2, leftBits) - 1)  puts(1, "   :  ")
? length(nodes)

puts(1,"Nodes used per input char                         :  ")
? length(nodes)/charCount


puts(1,"Total bytes for tree storage including pointers is:  ") 
t1 =  (bytes(nodes)  + bytes(rPointers))
? t1

puts(1,"Num of storage bytes per input char is            :  ")
? t1 / charCount
puts(1,"\n")

puts(1,"\nRNodes total limit : used                         :  " )
print(1, power(2, rightBits) - 1)  puts(1, " : ")
print(1,length(rPointers))

puts(1,"\nBytes to store RPointers)                         :  ")
? bytes(rPointers)
puts(1,"\n")

puts(1,"Bytes for inputWords Sequence, just for comparison:  ")
? wordsInBytes

puts(1,"Compression ratio bytes(sequence) / bytes(nodes)  :  ")
? wordsInBytes / bytes(nodes)
--cprint(rPointers)


abort(1)

--The following provide for tree saving and reloading but have not been
--completed...sort of waiting for a faster edo-load.

puts(1,"\nNodes before saving, length & numBytes is      : ")
printf(1,"%6d\t", length(nodes))  
? (bytes(nodes)+ bytes(rPointers))


puts(1,"Time for save by print()          : ")
fOut = open(outFileName1, "wb")
t1 = time()
print(fOut,nodes)
t2 = time()
printf(1,"%3.4f\t",t2-t1)

close(fOut)
puts(1,"Seconds \nTime for load by get()            : ")
nodes = {}
t1 = time()
fOut = open(outFileName1, "rb")
nodes = get(fOut)
t2 = time()
printf(1,"%3.4f\tSeconds",t2-t1)




puts(1,"Seconds\nTime for save by edo_save()       : ")
t3 = time()
o1 =  edo_save(outFileName2,nodes,CP_NONE)
t4 = time()
printf(1,"%3.4f\t",t4-t3)


puts(1,"Seconds \nTime for load by edoLoad()        : ")
nodes = {}
t1 = time()
nodes = edo_load(outFileName2,CP_NONE)
t2 = time()
printf(1,"%3.4f\t",t2-t1)
puts(1,"Seconds \n ")

puts(1,"\nNodes after edo_load, length & numBytes is : ")
printf(1,"%6d\t",length(nodes[2]))  ? bytes(nodes[2])

puts(1,"\nNodes after get, length & numBytes is      : ")
printf(1,"%6d\t", length(nodes[2]))  ? bytes(nodes[2])
